'''Define the Calculator class.'''


__author__ = 'Chao Wu'


from collections import ChainMap
from collections.abc import Iterable
from functools import reduce
from copy import deepcopy
import numpy as np
from numpy.random import normal
import pandas as pd
from scipy.linalg import null_space, pinv2, expm
from sympy import symbols, lambdify, Matrix, derive_by_array
try:
    import jax.numpy as jnp
    from jax import config, jacfwd
    config.update('jax_platform_name', 'cpu')
except ModuleNotFoundError:
    JAX_INSTALLED = False
else:
    JAX_INSTALLED = True
from multiprocessing import Pool
from ..core.mdv import MDV, get_natural_MDV, get_substrate_MDV, conv, diff_conv
import warnings
warnings.filterwarnings('ignore', category = RuntimeWarning)
warnings.filterwarnings('ignore', category = DeprecationWarning)

    
class Calculator():
    '''
    Parameters
    ----------
    model: Model
        Freeflux Model.
    '''
    
    def __init__(self, model):
        '''
        Parameters
        ----------
        model: Model
            Freeflux Model.
        '''
        
        self.model = model
        
    
    def _set_timepoints(self):
        
        ts = []
        for _, tpoint_infos in self.model.measured_inst_MDVs.items():
            ts.extend(tpoint_infos.keys())
        ts = sorted(set(ts))
        
        for t in ts:
            self.model.timepoints.append(t)
            
    
    def _calculate_null_space(self):
        
        S = self.model.get_total_stoichiometric_matrix(self.model.unbalanced_metabolites)
        self.model.null_space = null_space(S.values)
    
    
    def _calculate_transform_matrix(self):
        
        transMat = pd.DataFrame(
            0, 
            index = self.model.netfluxids, 
            columns = self.model.totalfluxids
        )
        for rxnid, rxn in self.model.reactions_info.items():
        
            if rxn.reversible:
                transMat.loc[rxnid, rxnid+'_f'] = 1
                transMat.loc[rxnid, rxnid+'_b'] = -1
            else:
                transMat.loc[rxnid, rxnid] = 1
        
        self.model.transform_matrix = transMat.values
    
    
    def _calculate_substrate_MDVs(self, extra_subs):
        
        if extra_subs is None:
             extra_subs = []

        for size in self.model.matrix_Bs:
            for sourceEMU in self.model.matrix_Bs[size][2]:
                
                if not isinstance(sourceEMU, Iterable):
                    sourceEMU = (sourceEMU,)
                    
                for emu in sourceEMU:
                    metabid = emu.metabolite_id
                    if metabid in self.model.end_substrates + extra_subs:
                        
                        if metabid in self.model.labeling_strategy:
                            atom_nos = emu.atom_nos
                            (labeling_pattern, 
                             percentage, 
                             purity) = self.model.labeling_strategy[metabid]
                            self.model.substrate_MDVs[emu] = get_substrate_MDV(
                                atom_nos, 
                                labeling_pattern, 
                                percentage, 
                                purity,
                                label_atom=self.model.label_atom
                            )
                        else:
                            natoms = emu.size
                            self.model.substrate_MDVs[emu] = get_natural_MDV(
                                natoms, 
                                base_atom=self.model.label_atom
                            )                        
        
    
    def _calculate_substrate_MDV_derivatives_basic(self, nvars, extra_subs):
        
        if extra_subs is None:
             extra_subs = []

        substrate_MDVs_der = {}
        for size in self.model.matrix_Bs:
            for sourceEMU in self.model.matrix_Bs[size][2]:
                
                if not isinstance(sourceEMU, Iterable):
                    sourceEMU = (sourceEMU,)
                    
                for emu in sourceEMU:
                    metabid = emu.metabolite_id
                    if metabid in self.model.end_substrates + extra_subs:
                        substrate_MDVs_der[emu] = np.zeros((emu.size+1, nvars))
                        
        return substrate_MDVs_der                
                        
                        
    def _calculate_substrate_MDV_derivatives_u(self, extra_subs):
        
        substrate_MDVs_der = self._calculate_substrate_MDV_derivatives_basic(
            self.model.null_space.shape[1], 
            extra_subs
        )
        
        return substrate_MDVs_der


    def _calculate_substrate_MDV_derivatives_c(self, extra_subs):
       
        substrate_MDVs_der = self._calculate_substrate_MDV_derivatives_basic(
            len(self.model.concids), 
            extra_subs
        )
        
        return substrate_MDVs_der
        
        
    def _calculate_substrate_MDV_derivatives_p(self, kind, extra_subs = None):
        '''
        Parameters
        ----------
        kind: {"ss", "inst"}
            * "ss" if isotopic steady state.
            * "inst" if isotopically nonstationary state.
        extra_subs: list or None
        '''
        
        if kind == 'ss':
            substrate_MDVs_der_u = self._calculate_substrate_MDV_derivatives_u(extra_subs)
            for emu in substrate_MDVs_der_u:
                self.model.substrate_MDVs_der_p[emu] = substrate_MDVs_der_u[emu]

        elif kind == 'inst':
            substrate_MDVs_der_u = self._calculate_substrate_MDV_derivatives_u(extra_subs)
            substrate_MDVs_der_c = self._calculate_substrate_MDV_derivatives_c(extra_subs)
            for emu in substrate_MDVs_der_u:
                self.model.substrate_MDVs_der_p[emu] = np.concatenate(
                    (substrate_MDVs_der_u[emu], substrate_MDVs_der_c[emu]), 
                    axis = 1
                )
    
    
    def _calculate_measured_fluxes_inversed_covariance_matrix(self):
        
        sds = np.array([sd for _, sd in self.model.measured_fluxes.values()])
        self.model.measured_fluxes_inv_cov = np.diag(1/sds**2) 
    
   
    def _calculate_measured_fluxes_derivative_v(self):
        
        measfluxids = list(self.model.measured_fluxes.keys())
        measured_fluxes_der = np.array(
            Matrix(measfluxids).jacobian(Matrix(self.model.totalfluxids))
        ).astype(float)
        
        return measured_fluxes_der
        
        
    def _calculate_measured_fluxes_derivative_c(self):
        
        nmeasfluxes = len(self.model.measured_fluxes)
        nmetabs = len(self.model.concids)
        measured_fluxes_der = np.zeros((nmeasfluxes, nmetabs))
        
        return measured_fluxes_der
        
        
    def _calculate_measured_fluxes_derivative_p(self, kind):
        '''
        Parameters
        ----------
        kind: {"ss", "inst"}
            * "ss" if isotopic steady state.
            * "inst" if isotopically nonstationary state.
        '''
        
        if kind == 'ss':
            measured_fluxes_der_v = self._calculate_measured_fluxes_derivative_v()
            measured_fluxes_der_u = measured_fluxes_der_v@self.model.null_space
            self.model.measured_fluxes_der_p = measured_fluxes_der_u

        elif kind == 'inst':
            measured_fluxes_der_v = self._calculate_measured_fluxes_derivative_v()
            measured_fluxes_der_u = measured_fluxes_der_v@self.model.null_space
            measured_fluxes_der_c = self._calculate_measured_fluxes_derivative_c()
            self.model.measured_fluxes_der_p = np.concatenate(
                (measured_fluxes_der_u, measured_fluxes_der_c), 
                axis = 1
            )

    
    def _generate_random_fluxes(self):
        
        self.ori_measured_fluxes = deepcopy(self.model.measured_fluxes)
        
        for fluxid, [mean, sd] in self.model.measured_fluxes.items():
            while True:
                meanNew = normal(mean, sd)
                if mean-3*sd <= meanNew <= mean+3*sd:
                    break
            self.model.measured_fluxes[fluxid][0] = meanNew
        
        
    def _reset_measured_fluxes(self):
        
        self.model.measured_fluxes = deepcopy(self.ori_measured_fluxes)
        
    
    def _calculate_measured_MDVs_inversed_covariance_matrix(self):
        
        sds = np.concatenate([sd for _, sd in self.model.measured_MDVs.values()])
        self.model.measured_MDVs_inv_cov = np.diag(1/sds**2)
        
        
    def _calculate_measured_inst_MDVs_inversed_covariance_matrix(self):
        
        sds = []
        for _, t_mdvs in self.model.measured_inst_MDVs.items():
            for t in sorted(t_mdvs):
                if t != 0:
                    sds.append(t_mdvs[t][1])
        sds = np.concatenate(sds)    
        
        self.model.measured_inst_MDVs_inv_cov = np.diag(1/sds**2)
        
    
    def _generate_random_MDVs(self):
        
        self.ori_measured_MDVs = deepcopy(self.model.measured_MDVs)
        
        for emuid, [means, sds] in self.model.measured_MDVs.items():
            mdv = MDV(normal(means, sds))
            self.model.measured_MDVs[emuid][0] = mdv.value


    def _reset_measured_MDVs(self):
        
        self.model.measured_MDVs = deepcopy(self.ori_measured_MDVs)
        

    def _generate_random_inst_MDVs(self):
        
        self.ori_measured_inst_MDVs = deepcopy(self.model.measured_inst_MDVs)

        for emuid, instMDVs in self.model.measured_inst_MDVs.items():
            for t, [means, sds] in instMDVs.items():
                if t != 0:
                    mdv = MDV(normal(means, sds))
                    self.model.measured_inst_MDVs[emuid][t][0] = mdv.value


    def _reset_measured_inst_MDVs(self):

        self.model.measured_inst_MDVs = deepcopy(self.ori_measured_inst_MDVs)
    

    @staticmethod
    def _calculate_matrix_A_and_B_derivatives_v(EAM, fluxids):
        '''
        Parameters
        ----------
        EAMstr: df
            EMU adjacency matrix of some size.
        fluxids: list
            Total fluxes IDs.
            
        Returns
        -------
        matrix_A_der, matrix_B_der: 3-D array
            Derivatives of A(B) w.r.t. total fluxes in shape of 
            (len(total fluxes), A(B).shape[0], A(B).shape[1]).
        '''
        
        import platform
        if platform.system() == 'Linux':
            import os
            os.sched_setaffinity(os.getpid(), range(os.cpu_count()))
            
        preAB = EAM.copy(deep = 'all').values
        colSum = preAB.sum(axis = 0)
        for i in range(colSum.size):
            preAB[i, i] = -colSum[i]
        
        A = preAB[:preAB.shape[1], :].T
        B = -preAB[preAB.shape[1]:, :].T
        
        matA = Matrix(A)
        matB = Matrix(B)
        
        if JAX_INSTALLED:
            lambA = lambdify(symbols(fluxids), matA, modules = 'jax')
            lambB = lambdify(symbols(fluxids), matB, modules = 'jax')
            
            matrix_A_der = np.array(
                jacfwd(lambA, range(len(fluxids)))(*jnp.ones(len(fluxids)))
            )
            matrix_B_der = np.array(
                jacfwd(lambB, range(len(fluxids)))(*jnp.ones(len(fluxids)))
            )
        else:
            matrix_A_der = np.array(
                derive_by_array(matA, symbols(fluxids)), 
                dtype = float
            )
            matrix_B_der = np.array(
                derive_by_array(matB, symbols(fluxids)), 
                dtype = float
            )
        
        return matrix_A_der, matrix_B_der
    
    
    def _calculate_matrix_As_and_Bs_derivatives_u(self, n_jobs):
        '''
        Parameters
        ----------
        n_jobs: int
            # of jobs to run in parallel.

        Returns
        -------
        matrix_A_der: 3-D array
            Derivatives of A(B) w.r.t. free fluxes in shape of 
            (len(free fluxes), A(B).shape[0], A(B).shape[1]).
        matrix_B_der: 3-D array
            Derivatives of A(B) w.r.t. free fluxes in shape of 
            (len(free fluxes), A(B).shape[0], A(B).shape[1]).        
        '''
            
        matrix_ABs_der = {}
        if n_jobs == 1:
            for size, EAM in self.model.EAMs.items():
                ABder = self._calculate_matrix_A_and_B_derivatives_v(EAM, self.model.totalfluxids)
                matrix_ABs_der[size] = ABder
        else:
            pool = Pool(processes = n_jobs)
            
            for size, EAM in self.model.EAMs.items():
                ABder = pool.apply_async(
                    func = self._calculate_matrix_A_and_B_derivatives_v, 
                    args = (EAM, self.model.totalfluxids)
                )
                matrix_ABs_der[size] = ABder
            
            pool.close()    
            pool.join()
            
            matrix_ABs_der = {size: ABder.get() for size, ABder in matrix_ABs_der.items()}
        
        matrix_As_der = {}
        matrix_Bs_der = {}
        for size, ABder in matrix_ABs_der.items():
            Ader = ABder[0].swapaxes(0,1).swapaxes(1,2)
            Bder = ABder[1].swapaxes(0,1).swapaxes(1,2)
            
            Ader = Ader@self.model.null_space
            Bder = Bder@self.model.null_space
        
            Ader = Ader.swapaxes(1,2).swapaxes(0,1)
            Bder = Bder.swapaxes(1,2).swapaxes(0,1)
            
            matrix_As_der[size] = Ader
            matrix_Bs_der[size] = Bder
            
        return matrix_As_der, matrix_Bs_der
        
            
    def _calculate_matrix_As_and_Bs_derivatives_c(self):
        '''
        Returns
        -------
        matrix_A_der: 3-D array
            Derivatives of A(B) w.r.t. total fluxes in shape of 
            (len(concs), A(B).shape[0], A(B).shape[1]).
        matrix_B_der: 3-D array
            Derivatives of A(B) w.r.t. total fluxes in shape of 
            (len(concs), A(B).shape[0], A(B).shape[1]).    
        '''
        
        nmetabs = len(self.model.concids)
        
        matrix_As_der = {}
        matrix_Bs_der = {}
        for size, EAM in self.model.EAMs.items():
            nEMUsout = EAM.shape[1]
            nEMUsin = EAM.shape[0] - nEMUsout
            
            matrix_As_der[size] = np.zeros((nmetabs, nEMUsout, nEMUsout))
            matrix_Bs_der[size] = np.zeros((nmetabs, nEMUsout, nEMUsin))
        
        return matrix_As_der, matrix_Bs_der
        
        
    def _calculate_matrix_As_and_Bs_derivatives_p(self, kind, n_jobs):
        '''
        Parameters
        ----------
        kind: {"ss", "inst"}
            * "ss" if isotopic steady state.
            * "inst" if isotopically nonstationary state.
        n_jobs: int
            # of jobs to run in parallel.
        '''
        
        if kind == 'ss':
            (matrix_As_der_u, 
             matrix_Bs_der_u
            ) = self._calculate_matrix_As_and_Bs_derivatives_u(n_jobs)
            for size in self.model.EAMs:
                self.model.matrix_As_der_p[size] = matrix_As_der_u[size]
                self.model.matrix_Bs_der_p[size] = matrix_Bs_der_u[size]

        elif kind == 'inst':
            (matrix_As_der_u, 
             matrix_Bs_der_u
            ) = self._calculate_matrix_As_and_Bs_derivatives_u(n_jobs)
            (matrix_As_der_c, 
             matrix_Bs_der_c
            ) = self._calculate_matrix_As_and_Bs_derivatives_c()
            for size in self.model.EAMs:
                self.model.matrix_As_der_p[size] = np.concatenate(
                    (matrix_As_der_u[size], matrix_As_der_c[size]), 
                    axis = 0
                )
                self.model.matrix_Bs_der_p[size] = np.concatenate(
                    (matrix_Bs_der_u[size], matrix_Bs_der_c[size]), 
                    axis = 0
                )
    
    
    def _lambdify_matrix_As_and_Bs(self):
       
        for size, EAM in self.model.EAMs.items():
            
            preAB = EAM.copy(deep = 'all')
            for emu in EAM.columns:
                preAB.loc[emu, emu] = -preAB[emu].sum() 
        
            A = preAB.loc[preAB.columns, :].T
            B = -preAB.loc[preAB.index.difference(preAB.columns), :].T
            
            matA = Matrix(A)
            matB = Matrix(B)
            
            fluxidsA = list(map(str, matA.free_symbols))
            fluxidsB = list(map(str, matB.free_symbols))
            
            lambA = lambdify(fluxidsA, matA, modules = 'numpy')
            lambB = lambdify(fluxidsB, matB, modules = 'numpy')
            
            self.model.matrix_As[size] = [lambA, fluxidsA, A.columns.tolist()]
            self.model.matrix_Bs[size] = [lambB, fluxidsB, B.columns.tolist()]
    
    
    def _calculate_matrix_Ms_derivatives_u(self):
        
        nfreefluxes = self.model.null_space.shape[1]
        
        matrix_Ms_der = {}
        for size, EAM in self.model.EAMs.items():
            nEMUsout = EAM.shape[1]
            matrix_Ms_der[size] = np.zeros((nfreefluxes, nEMUsout, nEMUsout))
        
        return matrix_Ms_der
        
    
    def _calculate_matrix_Ms_derivatives_c(self):
        
        matrix_Ms_der = {}
        for size, EAM in self.model.EAMs.items():            
            
            matM = Matrix(np.diag(symbols([emu.metabolite_id for emu in EAM.columns])))
            if JAX_INSTALLED:
                lambM = lambdify(symbols(self.model.concids), matM, modules = 'jax')
                matrix_M_der = np.array(
                    jacfwd(lambM, 
                           range(len(self.model.concids))
                    )(*jnp.ones(len(self.model.concids)))
                )
            else:
                matrix_M_der = np.array(
                    derive_by_array(matM, symbols(self.model.concids)), 
                    dtype = float
                )
            matrix_Ms_der[size] = matrix_M_der
            
        return matrix_Ms_der
        
        
    def _calculate_matrix_Ms_derivatives_p(self):
        
        matrix_Ms_der_u = self._calculate_matrix_Ms_derivatives_u()
        matrix_Ms_der_c = self._calculate_matrix_Ms_derivatives_c()
        
        for size in self.model.EAMs:
            self.model.matrix_Ms_der_p[size] = np.concatenate(
                (matrix_Ms_der_u[size], matrix_Ms_der_c[size]), 
                axis = 0
            )
        
    
    def _lambdify_matrix_Ms(self):
        
        for size, EAM in self.model.EAMs.items():
            matM = Matrix(np.diag(symbols([emu.metabolite_id for emu in EAM.columns])))
            metabids = list(map(str, matM.free_symbols))
            lambM = lambdify(metabids, matM, modules = 'numpy')
            self.model.matrix_Ms[size] = [lambM, metabids]
        
    
    def _calculate_initial_matrix_Xs(self):
        
        for size in self.model.matrix_As:
            nEMUs = len(self.model.matrix_As[size][2])
            iniX = np.vstack(
                [get_natural_MDV(size, base_atom=self.model.label_atom).value]*nEMUs
            )
            self.model.initial_matrix_Xs[size] = iniX
            
            
    def _calculate_initial_matrix_Ys(self):
        
        for size in self.model.matrix_Bs:
            iniY = []
            for sourceEMU in self.model.matrix_Bs[size][2]:
                if not isinstance(sourceEMU, Iterable):
                    sourceMDV = self.model.substrate_MDVs[sourceEMU]
                else:
                    mdvs = []
                    for emu in sourceEMU:
                        if emu not in self.model.substrate_MDVs:
                            mdv = get_natural_MDV(
                                emu.size, 
                                base_atom=self.model.label_atom
                            )
                        else:
                            mdv = self.model.substrate_MDVs[emu]
                        mdvs.append(mdv)
                    sourceMDV = reduce(conv, mdvs)
                iniY.append(sourceMDV)
            iniY = np.array(iniY)
            self.model.initial_matrix_Ys[size] = iniY
        
    
    def _calculate_initial_matrix_Xs_derivatives_p(self):
        
        nvars = self.model.null_space.shape[1] + len(self.model.concids)
        for size, iniX in self.model.initial_matrix_Xs.items():
            Xshape = iniX.shape
            iniXder = np.zeros((nvars, *Xshape))
            self.model.initial_matrix_Xs_der_p[size] = iniXder
        
        
    def _calculate_initial_matrix_Ys_derivatives_p(self):
        
        nvars = self.model.null_space.shape[1] + len(self.model.concids)
        for size, iniY in self.model.initial_matrix_Ys.items():
            Yshape = iniY.shape
            iniYder = np.zeros((nvars, *Yshape))
            self.model.initial_matrix_Ys_der_p[size] = iniYder
        
    
    def _build_initial_sim_MDVs(self):
        
        for size in sorted(self.model.matrix_As):
            productEMUs = self.model.matrix_As[size][2]
            iniX = self.model.initial_matrix_Xs[size]        
            for productEMU, iniMDV in zip(productEMUs, iniX):
                if productEMU.id in self.model.target_EMUs:
                    self.model.initial_sim_MDVs[productEMU.id] = {0: MDV(iniMDV)}
                    
        
    def _calculate_MDVs(self):
        '''
        This method simulate MDVs at isotopically steady state.
        
        Returns
        -------
        simMDVs: dict
            EMU ID => MDV (in array).
        '''
        
        simMDVs = {}
        for size in sorted(self.model.matrix_As):
            
            lambA, fluxidsA, productEMUs = self.model.matrix_As[size]
            lambB, fluxidsB, sourceEMUs = self.model.matrix_Bs[size]
            
            A = lambA(*self.model.total_fluxes[fluxidsA])
            B = lambB(*self.model.total_fluxes[fluxidsB])
            
            Y = []
            for sourceEMU in sourceEMUs:
                if not isinstance(sourceEMU, Iterable):
                    sourceMDV = self.model.substrate_MDVs[sourceEMU]
                else:
                    mdvs = [ChainMap(simMDVs, self.model.substrate_MDVs)[emu] 
                            for emu in sourceEMU]
                    sourceMDV = reduce(conv, mdvs)
                Y.append(sourceMDV)    
            Y = np.array(Y)
            
            X = pinv2(A, check_finite = True)@B@Y
        
            simMDVs.update(zip(productEMUs, X))
            
        simMDVs = {emu.id: mdv for emu, mdv in simMDVs.items()}
        
        return simMDVs
                    
    
    def _calculate_MDVs_and_derivatives_p(self):
        '''
        This method simulate MDVs and their derivatives at isotopically steady state.
        
        Returns
        -------
        simMDVs: dict
            EMU ID => MDV (in array).
        simMDVsDer: dict
            EMU ID => 2-D array in shape of (len(MDV), len(u)).
        '''
        
        simMDVs = {}
        simMDVsDer = {}
        for size in sorted(self.model.matrix_As):
            
            lambA, fluxidsA, productEMUs = self.model.matrix_As[size]
            lambB, fluxidsB, sourceEMUs = self.model.matrix_Bs[size]
            
            A = lambA(*self.model.total_fluxes[fluxidsA])
            B = lambB(*self.model.total_fluxes[fluxidsB])
            
            Ainv = pinv2(A, check_finite = True)
            
            Ader = self.model.matrix_As_der_p[size]
            Bder = self.model.matrix_Bs_der_p[size]
            
            Y = []
            Yder = []
            for sourceEMU in sourceEMUs:
                if not isinstance(sourceEMU, Iterable):
                    sourceMDV = self.model.substrate_MDVs[sourceEMU]
                    sourceMDVder = self.model.substrate_MDVs_der_p[sourceEMU]
                else:
                    mdvs = []
                    mdvs_mdvders = []
                    for emu in sourceEMU:
                        mdv = ChainMap(simMDVs, self.model.substrate_MDVs)[emu]
                        mdvs.append(mdv)
                        mdvder = ChainMap(simMDVsDer, self.model.substrate_MDVs_der_p)[emu]
                        mdvs_mdvders.append([mdv, mdvder])
                    sourceMDV = reduce(conv, mdvs)
                    sourceMDVder = reduce(diff_conv, mdvs_mdvders)[1]
                Y.append(sourceMDV)
                Yder.append(sourceMDVder)
            Y = np.array(Y)
            Yder = np.array(Yder).swapaxes(1,2).swapaxes(0,1)
            
            X = Ainv@B@Y
            Xder = Ainv@(Bder@Y + B@Yder - Ader@X)
            Xder = Xder.swapaxes(0,1).swapaxes(1,2)
            
            simMDVs.update(zip(productEMUs, X))
            simMDVsDer.update(zip(productEMUs, Xder))
            
        simMDVs = {emu.id: mdv for emu, mdv in simMDVs.items()}
        simMDVsDer = {emu.id: mdvDer for emu, mdvDer in simMDVsDer.items()}
        
        return simMDVs, simMDVsDer
    
    
    def _calculate_inst_MDVs(self):
        '''
        This method simulate MDVs at isotopically nonstationary state.
        
        Returns
        -------
        simInstMDVs: dict
            EMU ID => {t => MDV (in array)} (starting from t1).
        '''
        
        simInstMDVs = {}
        Ys = {}
        Xs = {}

        t1 = 0.0
        for size in sorted(self.model.matrix_As):
            Y_t1 = self.model.initial_matrix_Ys[size]
            Ys.setdefault(t1, {})[size] = Y_t1
            
            X_t1 = self.model.initial_matrix_Xs[size]
            Xs.setdefault(t1, {})[size] = X_t1

        for t in self.model.timepoints:
            if t != 0.0:
                t0 = t1
                t1 = t            
                deltat = t1 - t0
            
                for size in sorted(self.model.matrix_As):
                    lambA, fluxidsA, productEMUs = self.model.matrix_As[size]
                    lambB, fluxidsB, sourceEMUs = self.model.matrix_Bs[size]
                    lambM, metabids =  self.model.matrix_Ms[size]
                    
                    A = lambA(*self.model.total_fluxes[fluxidsA])
                    B = lambB(*self.model.total_fluxes[fluxidsB])
                    M = lambM(*self.model.concentrations[metabids])
                    Minv = pinv2(M, check_finite = True)
                    
                    F = Minv@A
                    Finv = pinv2(F, check_finite = True)
                    I = np.eye(*F.shape)
                    Phi = expm(F*deltat)
                    Gamma = (Phi - I)@Finv
                    Omega = (Gamma/deltat - I)@Finv
                    
                    X_t0 = Xs[t0][size]
                    
                    Y_t0 = Ys[t0][size]
                    G_t0 = Minv@B@Y_t0
                    
                    Y_t1 = []
                    for sourceEMU in sourceEMUs:
                        if not isinstance(sourceEMU, Iterable):
                            sourceMDV = self.model.substrate_MDVs[sourceEMU]
                        else:
                            mdvs = []
                            for emu in sourceEMU:
                                mdv = ChainMap(simInstMDVs, self.model.substrate_MDVs)[emu]
                                if isinstance(mdv, dict):
                                    mdv = mdv[t1]
                                mdvs.append(mdv)
                            sourceMDV = reduce(conv, mdvs)
                        Y_t1.append(sourceMDV)    
                    Y_t1 = np.array(Y_t1)
                    G_t1 = Minv@B@Y_t1
                    
                    X_t1 = Phi@X_t0 - Gamma@G_t0 - Omega@(G_t1 - G_t0)
                    
                    Ys.setdefault(t1, {})[size] = Y_t1
                    Xs.setdefault(t1, {})[size] = X_t1
                    
                    for productEMU, mdv_t1 in zip(productEMUs, X_t1):
                        simInstMDVs.setdefault(productEMU, {}).update({t1: mdv_t1})
                    
        simInstMDVs = {emu.id: mdvs for emu, mdvs in simInstMDVs.items()}
        
        return simInstMDVs        


    def _calculate_inst_MDVs_and_derivatives_p(self):
        '''
        This method simulate MDVs and their derivatives at isotopically nonstationary state.
        
        Returns
        -------
        simInstMDVs: dict
            EMU ID => {t => MDV (in array)} (starting from t1).
        simInstMDVsDer: dict
            EMU ID => {t => 2-D array in shape of (len(MDV), len(u)+len(c))} (starting from t1).
        '''
        
        simInstMDVs = {}
        simInstMDVsDer = {}
        Ys = {}
        Xs = {}
        Yders = {}
        Xders = {}
        
        t1 = 0.0
        for size in sorted(self.model.matrix_As):
            Y_t1 = self.model.initial_matrix_Ys[size]
            Ys.setdefault(t1, {})[size] = Y_t1
            
            X_t1 = self.model.initial_matrix_Xs[size]
            Xs.setdefault(t1, {})[size] = X_t1
            
            Yder_t1 = self.model.initial_matrix_Ys_der_p[size]
            Yders.setdefault(t1, {})[size] = Yder_t1
            
            Xder_t1 = self.model.initial_matrix_Xs_der_p[size]
            Xders.setdefault(t1, {})[size] = Xder_t1
        
        for t in self.model.timepoints:
            if t != 0.0:
                t0 = t1
                t1 = t            
                deltat = t1 - t0
            
                for size in sorted(self.model.matrix_As):
                    lambA, fluxidsA, productEMUs = self.model.matrix_As[size]
                    lambB, fluxidsB, sourceEMUs = self.model.matrix_Bs[size]
                    lambM, metabids =  self.model.matrix_Ms[size]
            
                    A = lambA(*self.model.total_fluxes[fluxidsA])
                    B = lambB(*self.model.total_fluxes[fluxidsB])
                    M = lambM(*self.model.concentrations[metabids])
                    Minv = pinv2(M, check_finite = True)
                    
                    Ader = self.model.matrix_As_der_p[size]
                    Bder = self.model.matrix_Bs_der_p[size]
                    Mder = self.model.matrix_Ms_der_p[size]
                    Minvder = -Minv@Mder@Minv
                    
                    F = Minv@A
                    Finv = pinv2(F, check_finite = True)
                    I = np.eye(*F.shape)
                    Phi = expm(F*deltat)
                    Gamma = (Phi - I)@Finv
                    Omega = (Gamma/deltat - I)@Finv
                    
                    X_t0 = Xs[t0][size]
                    
                    Y_t0 = Ys[t0][size]
                    G_t0 = Minv@B@Y_t0
                    
                    Xder_t0 = Xders[t0][size]
                    
                    Yder_t0 = Yders[t0][size]
                    H_t0 = (Minvder@A@X_t0 
                            + Minv@Ader@X_t0 
                            - Minv@B@Yder_t0 
                            - Minvder@B@Y_t0 
                            - Minv@Bder@Y_t0)
                    
                    Y_t1 = []
                    Yder_t1 = []
                    for sourceEMU in sourceEMUs:
                        if not isinstance(sourceEMU, Iterable):
                            sourceMDV = self.model.substrate_MDVs[sourceEMU]
                            sourceMDVder = self.model.substrate_MDVs_der_p[sourceEMU]
                        else:
                            mdvs = []
                            mdvs_mdvders = []
                            for emu in sourceEMU:
                                mdv = ChainMap(simInstMDVs, self.model.substrate_MDVs)[emu]
                                if isinstance(mdv, dict): 
                                    mdv = mdv[t1]
                                mdvs.append(mdv)    
                                mdvder = ChainMap(simInstMDVsDer, self.model.substrate_MDVs_der_p)[emu]
                                if isinstance(mdvder, dict): 
                                    mdvder = mdvder[t1]
                                mdvs_mdvders.append([mdv, mdvder])
                            sourceMDV = reduce(conv, mdvs)
                            sourceMDVder = reduce(diff_conv, mdvs_mdvders)[1]
                        Y_t1.append(sourceMDV)
                        Yder_t1.append(sourceMDVder)
                    Y_t1 = np.array(Y_t1)
                    Yder_t1 = np.array(Yder_t1).swapaxes(1,2).swapaxes(0,1)
                    
                    G_t1 = Minv@B@Y_t1
                    X_t1 = Phi@X_t0 - Gamma@G_t0 - Omega@(G_t1 - G_t0)
                    
                    H_t1 = (Minvder@A@X_t1 
                            + Minv@Ader@X_t1 
                            - Minv@B@Yder_t1 
                            - Minvder@B@Y_t1 
                            - Minv@Bder@Y_t1)
                    Xder_t1 = Phi@Xder_t0 + Gamma@H_t0 + Omega@(H_t1 - H_t0)
                    
                    Ys.setdefault(t1, {})[size] = Y_t1
                    Xs.setdefault(t1, {})[size] = X_t1
                    Yders.setdefault(t1, {})[size] = Yder_t1
                    Xders.setdefault(t1, {})[size] = Xder_t1
                    
                    for productEMU, mdv_t1 in zip(productEMUs, X_t1):
                        simInstMDVs.setdefault(productEMU, {}).update({t1: mdv_t1})
                        
                    Xder_t1 = Xder_t1.swapaxes(0,1).swapaxes(1,2)    
                    for productEMU, mdvder_t1 in zip(productEMUs, Xder_t1):
                        simInstMDVsDer.setdefault(productEMU, {}).update({t1: mdvder_t1})
                        
        simInstMDVs = {emu.id: mdvs for emu, mdvs in simInstMDVs.items()}                
        simInstMDVsDer = {emu.id: mdvders for emu, mdvders in simInstMDVsDer.items()}
        
        return simInstMDVs, simInstMDVsDer



'''Define the Context class.'''


__author__ = 'Chao Wu'


class Context():
    
    def __init__(self):
        
        self.operations = []
        
    
    def add_undo(self, op):
        '''
        Parameters
        ----------
        op: callable
            Operation to reset the model.
        '''
        
        self.operations.append(op)


    def undo(self):
        
        while self.operations:
            op = self.operations.pop()
            op()



'''Define the InstSimulator class.'''


__author__ = 'Chao Wu'


from collections.abc import Iterable
from functools import partial
from ..core.mdv import MDV
from ..analysis.simulate import Simulator
from ..io.inputs import read_preset_values_from_file 
from ..io.results import InstSimResults


class InstSimulator(Simulator):
    '''
    Provided fluxes should be in the unit of umol/gCDW/s if concentrations in the unit of 
    umol/gCDW and timepoints in the unit of s.
    '''
    
    def set_concentration(self, metabid, value):
        '''
        Set metabolite concentration in unit of umol/gCDW.
        
        Parameters
        ----------
        metabid: str
            Metabolite ID.
        value: float
            Intracellular concentration.
        '''
        
        self.model.concentrations[metabid] = value
        
        if self.contexts:
            context = self.contexts[-1]
            context.add_undo(partial(self._unset_concentrations, metabid))
        
        
    def set_concentrations_from_file(self, file):
        '''
        Read metabolite concentrations (umol/gCDW) from file.
        
        Parameters
        ----------
        file: file path
            tsv or excel file with fields "metabolite_ID" and "value".
        '''
        
        concs = read_preset_values_from_file(file)
                
        for metabid, value in concs.items():
            self.model.concentrations[metabid] = value
        
        if self.contexts:
            context = self.contexts[-1]
            context.add_undo(
                partial(self._unset_concentrations, concs.index.tolist())
            )    

    
    def _unset_concentrations(self, metabids):
        '''
        Parameters
        ----------
        metabids: str or list of str
            Metabolite ID(s).
        '''
        
        if not isinstance(metabids, Iterable):
            metabids = [metabids]
        
        for metabid in metabids: 
            if metabid in self.model.concentrations:
                self.model.concentrations.drop(metabids, inplace = True)
                
                
    def set_timepoints(self, timepoints):
        '''
        Set timepoints in unit of s.
        
        Parameters
        timepoints: list of float
            Timepoints when MDVs are simulated.
        '''
        
        self.model.timepoints = sorted(set(self.model.timepoints + timepoints))
        if 0 not in self.model.timepoints:
            self.model.timepoints = [0] + self.model.timepoints
        
        if self.contexts:
            context = self.contexts[-1]
            context.add_undo(self._unset_timepoints)
        
    
    def _unset_timepoints(self):
        
        self.model.timepoints.clear()
    
    
    def _lambdify_matrix_Ms(self):
        
        if not self.model.matrix_Ms:
            self.calculator._lambdify_matrix_Ms()
        
        if self.contexts:
            context = self.contexts[-1]
            context.add_undo(self._unset_matrix_Ms)
        
        
    def _unset_matrix_Ms(self):
        
        self.model.matrix_Ms.clear()
            
            
    def _calculate_initial_matrix_Xs(self):
        
        if not self.model.initial_matrix_Xs:
            self.calculator._calculate_initial_matrix_Xs()
        
        if self.contexts:
            context = self.contexts[-1]
            context.add_undo(self._unset_initial_matrix_Xs)
        
        
    def _unset_initial_matrix_Xs(self):
        
        self.model.initial_matrix_Xs.clear()
        
        
    def _calculate_initial_matrix_Ys(self):
        
        if not self.model.initial_matrix_Ys:
            self.calculator._calculate_initial_matrix_Ys()
        
        if self.contexts:
            context = self.contexts[-1]
            context.add_undo(self._unset_initial_matrix_Ys)
            
            
    def _unset_initial_matrix_Ys(self):
        
        self.model.initial_matrix_Ys.clear()
        
        
    def _build_initial_sim_MDVs(self):
        
        if not self.model.initial_sim_MDVs:
            self.calculator._build_initial_sim_MDVs()
            
        if self.contexts:
            context = self.contexts[-1]
            context.add_undo(self._unset_initial_sim_MDVs)
            
            
    def _unset_initial_sim_MDVs(self):
        
        self.model.initial_sim_MDVs.clear()
        
    
    def prepare(self, n_jobs = 1):
        '''
        Parameters
        ----------
        n_jobs: int
            If n_jobs > 1, decomposition job will run in parallel.
        '''
        
        self._decompose_network(n_jobs, lump = False)
        self._lambdify_matrix_As_and_Bs()
        self._lambdify_matrix_Ms()        
        self._calculate_substrate_MDVs()
        self._calculate_initial_matrix_Xs()
        self._calculate_initial_matrix_Ys()
        self._build_initial_sim_MDVs()
        

    def _check_dependencies(self):

        if self.model.total_fluxes.empty:
            raise ValueError('call set_flux or set_fluxes_from_file first')
        if self.model.concentrations.empty:
            raise ValueError('call set_concentration or set_concentrations_from_file first')
        if not self.model.timepoints:
            raise ValueError('call set_timepoints first')
        if not self.model.target_EMUs:
            raise ValueError('call set_target_EMUs first')    
        if not self.model.labeling_strategy:
            raise ValueError('call labeling_strategy first')
        
        checklist = [
            not self.model.EAMs, 
            not self.model.substrate_MDVs, 
            not self.model.initial_matrix_Xs, 
            not self.model.matrix_As, 
            not self.model.matrix_Bs, 
            not self.model.matrix_Ms,
            not self.model.timepoints
        ]
        if any(checklist):
            raise ValueError('call prepare first')


    def simulate(self):
        '''
        Returns
        -------
        instTargetMDVs: dict of dict
           EMU IDs => {measurement timepoints => simulated MDVs}
        '''
        
        simInstMDVs = self.calculator._calculate_inst_MDVs()
        
        targetInstMDVs = {}
        for emuid in self.model.target_EMUs:
            instMDVs = self.model.initial_sim_MDVs[emuid].copy()
            instMDVs.update({t: MDV(simInstMDVs[emuid][t]) for t in simInstMDVs[emuid]})
            targetInstMDVs[emuid] = instMDVs
        
        return InstSimResults(targetInstMDVs)



'''Define the Reaction class.'''


__author__ = 'Chao Wu'


from collections import ChainMap, Counter
from itertools import product
from functools import reduce, lru_cache
import numpy as np
import pandas as pd
from sympy import Symbol
from .emu import EMU


class Reaction():
    '''
    Define Reaction object which constitute a Model.

    Duplicate substrates or products could appear in one reaction, but with different atoms.
    
    Parameters
    ----------
    id: str
        Reaction ID.
    reversible: bool
        Reversibility.

    Attributes
    ----------
    id: str
        Reaction ID.
    reversible: bool
        Reversibility.
    substrates_info: df
        Index are substrate IDs (there could be duplicate substrates), 
        columns are Metabolite object and its stoichiometric number.
    products_info: df
        Index are product IDs (there could be duplicate products), 
        columns are Metabolite object and stoichiometrc number.
    substrates: list
        Unique substrate IDs, in order of alphabet.
    products: list
        Unique product IDs, in order of alphabet.
    substrates_with_atoms: list
        Unique IDs of substrates with atoms, in order of alphabet.
    products_with_atoms: list    
        Unique IDs of products with atoms, in order of alphabet.
    _substrates_atom_mapping: list of dict or None
        For example, reactants like: A({'ab': 0.5, 'ba': 0.5}) + B({'c': 1}) will be transformed to
        [{'a': [A, 1, 0.5], 'b': [A, 2, 0.5], 'c': [C, 1, 1]},
         {'a': [A, 2, 0.5], 'b': [A, 1, 0.5], 'a': [C, 1, 1]}]
    _products_atom_mapping: list of dict or None
        For example, reactants like: A({'ab': 0.5, 'ba': 0.5}) + B({'c': 1}) will be transformed to
        [{'a': [A, 1, 0.5], 'b': [A, 2, 0.5], 'c': [C, 1, 1]},
         {'a': [A, 2, 0.5], 'b': [A, 1, 0.5], 'a': [C, 1, 1]}]    
    flux (fflux and bflux for reversible reaction): Symbol
        Reaction flux (forward flux and backward flux for reversible reaction).
    host_models: set of Model or None
        Model hosting the reaction.
    '''
    
    def __init__(self, id, reversible = True):
        '''
        Parameters
        ----------
        id: str
            Reaction ID.
        reversible: bool
            Reversibility.
        '''
        
        self.id = id
        self.reversible = reversible
        self.substrates_info = pd.DataFrame(columns = ['metab', 'stoy'])
        self.products_info = pd.DataFrame(columns = ['metab', 'stoy'])
        if self.reversible:
            self.fflux = Symbol(self.id+'_f')
            self.bflux = Symbol(self.id+'_b')
        else:
            self.flux = Symbol(self.id)
        self.host_models = None
    
    
    def add_substrates(self, substrates, stoichiometry):
        '''
        Parameters
        ----------
        substrates: Metabolite or list of Metabolite
        stoichiometry: float or list of float
            Stoichiometric number(s) of corresponding substrate(s).
        '''
        
        if not isinstance(substrates, list):
            substrates = [substrates]
            stoichiometry = [stoichiometry]
            
        newSubs = pd.DataFrame(
            {'metab': substrates, 'stoy': np.array(stoichiometry).astype(np.float)}, 
            index = [sub.id for sub in substrates]
        )    
            
        self.substrates_info = pd.concat((self.substrates_info, newSubs))
        
        
        for sub in substrates:
            if sub.host_reactions is None:
                sub.host_reactions = set([self])
            else:
                sub.host_reactions.add(self)
        

    def add_products(self, products, stoichiometry):
        '''
        Parameters
        ----------
        products: Metabolite or list of Metabolite.
        stoichiometry: float or list of float
            stoichiometric number(s) of corresponding product(s).
        '''
        
        if not isinstance(products, list):
            products = [products]
            stoichiometry = [stoichiometry]
            
        newPros = pd.DataFrame(
            {'metab': products, 'stoy': np.array(stoichiometry).astype(np.float)}, 
            index = [pro.id for pro in products]
        )    
            
        self.products_info = pd.concat((self.products_info, newPros))    
        
        
        for pro in products:
            if pro.host_reactions is None:
                pro.host_reactions = set([self])
            else:
                pro.host_reactions.add(self)
                

    def remove_substrates(self, substrates):
        '''
        Parameters
        ----------
        substrates: Metabolite or list of Metabolite
        '''
        
        if not isinstance(substrates, list):
            substrates = [substrates]
        
        self.substrates_info.drop([sub.id for sub in substrates], inplace = True)
        
        
        for sub in substrates:
            sub.host_reactions.remove(self)
            
            if not sub.host_reactions:
                sub.host_reactions = None
                break
        
    
    def remove_products(self, products):
        '''
        Parameters
        ----------
        products: Metabolite or list of Metabolite
        '''
        
        if not isinstance(products, list):
            products = [products]
        
        self.products_info.drop([pro.id for pro in products], inplace = True)
        
        
        for pro in products:
            pro.host_reactions.remove(self)
            
            if not pro.host_reactions:
                pro.host_reactions = None
                break
    
    
    @property
    @lru_cache()
    def substrates(self):
        
        return sorted(self.substrates_info.index.unique().tolist())
        
        
    @property
    @lru_cache()
    def products(self):
        
        return sorted(self.products_info.index.unique().tolist())
        
        
    @property
    @lru_cache()
    def substrates_with_atoms(self):
        
        return sorted({sub for sub, (metab, row) in self.substrates_info.iterrows() if metab.atoms})
        
        
    @property
    @lru_cache()
    def products_with_atoms(self):
        
        return sorted({pro for pro, (metab, row) in self.products_info.iterrows() if metab.atoms})    
        
    
    def _atom_mapping(self, reactant):
        '''
        Parameters
        ----------
        reactant: str,
            'substrate' or 'product'

        Notes
        -----
        Reactants like: A({'ab': 0.5, 'ba': 0.5}) + B({'c': 1}) will be transformed to
        [{'a': [A, 1, 0.5], 'b': [A, 2, 0.5], 'c': [C, 1, 1]},
         {'a': [A, 2, 0.5], 'b': [A, 1, 0.5], 'a': [C, 1, 1]}].
        '''
        
        if reactant == 'substrate':
            reacsInfo = self.substrates_info.loc[self.substrates_with_atoms, 'metab']
        elif reactant == 'product':
            reacsInfo = self.products_info.loc[self.products_with_atoms, 'metab']
        else:
            raise ValueError('must be "substrate" or "product"')
        
        reacsAtomInfo = []
        for _, Metab in reacsInfo.items():
            
            reacAtomInfo = []
            for atoms, coe in Metab.atoms_info.items():
                
                atomInfo = {atom: [Metab, no+1, coe] for no, atom in enumerate(atoms)}
                reacAtomInfo.append(atomInfo)
        
            reacsAtomInfo.append(reacAtomInfo)
        
        reacsAtomMappingRaw = list(product(*reacsAtomInfo))
        
        reacsAtomMapping = [ChainMap(*scenario) for scenario in reacsAtomMappingRaw]
        
        return reacsAtomMapping
        
    
    @property
    @lru_cache()
    def _substrates_atom_mapping(self):
        
        if self.substrates_with_atoms:
            return self._atom_mapping('substrate')
        else:
            return None
        
    
    @property
    @lru_cache()
    def _products_atom_mapping(self):
        
        if self.products_with_atoms:
            return self._atom_mapping('product')
        else:
            return None
            
    
    def _find_precursor_EMUs(self, emu, direction = 'forward'):
        '''
        Parameters
        ----------
        emu: EMU
        direction: str
            * For reversible reaction,
            'forward' if emu is product and precursor emu(s) are substrates;
            'backward' if emu is substrate and precursor emu(s) are products.
            * For irreversible reaction, only 'forward' is acceptable.
            
        Returns
        -------
        preEMUsInfo: list

        Notes
        -----
        For reaction like: A({'ab': 0.5, 'ba': 0.5}) + B({'c': 1}) -> C({'abc': 0.5, 'cba': 0.5}),
        _find_precursor_EMUs(C12) returns
        [[[A_12], 0.5],
         [[B_1, A_2], 0.25],
         [[B_1, A_1], 0.25]].
        '''        
        
        if self.reversible:
            if direction == 'forward':
                atomMapping = self._substrates_atom_mapping
                reacInfo = self.substrates_info['metab']
            
            elif direction == 'backward':
                atomMapping = self._products_atom_mapping
                reacInfo = self.products_info['metab']
        
        else:
            if direction != 'forward':
                raise ValueError('only "forward" is acceptable for irreversible reaction')
            
            atomMapping = self._substrates_atom_mapping
            reacInfo = self.substrates_info['metab']
        
        
        preEMUsInfoRaw = []
        for scenario in atomMapping:
            for atoms, coe in emu.metabolite.atoms_info.items():   
                
                preAtoms = {}
                uniCoe = coe
                for atom in [atoms[no-1] for no in emu.atom_nos]:
                    
                    pre, preAtomNO, preCoe = scenario[atom]   
                    
                    if pre not in preAtoms:
                        uniCoe *= preCoe
                        preAtoms[pre] = [preAtomNO]
                    else:
                        preAtoms[pre].append(preAtomNO)
                
                preEMUs = [EMU(pre.id+'_'+''.join(map(str, sorted(preAtomNOs))), pre, preAtomNOs) 
                           for pre, preAtomNOs in preAtoms.items()]
                preEMUsInfoRaw.append([preEMUs, uniCoe])
        
        preEMUsInfoRaw = [Counter({tuple(sorted(preEMUs)): coe}) 
                          for preEMUs, coe in preEMUsInfoRaw]   
        preEMUsInfo = reduce(lambda x, y: x+y, preEMUsInfoRaw) 
        preEMUsInfo = [[list(preEMUs), coe] for preEMUs, coe in preEMUsInfo.items()]
        
        return preEMUsInfo
        
    
    def __repr__(self):
        
        arrow = '<->' if self.reversible else '->'
        subsStr = '+'.join(self.substrates)
        prosStr = '+'.join(self.products)
        
        return f'{self.__class__.__name__} {self.id}: {subsStr}{arrow}{prosStr}'        



'''Define the Metabolite class.'''


__author__ = 'Chao Wu'


import re


class Metabolite():
    '''
    Define Metabolite object which constitutes a Reaction.

    Metabolites are considered identical when they have the same IDs and atoms.
    
    Parameters
    ----------
    id: str
        Metabolite ID.
    atoms: str or list of str or None
        Carbons in metabolite, e.g., 'abcd' for metabolite without equivalents, 
        and ['abcd', 'dcba'] for metabolite with equivalents.

    Attributes
    ----------
    id: str
        Metabolite ID.
    atoms_info: dict
        equivalent => coefficient, e.g., {'abcd': 1.0} or {'abcd': 0.5, 'dcba': 0.5}.
    n_carbons: int
        # of carbons in metabolite. 
    host_reactions: set of Reaction or None
        Reactions hosting the metabolite.
    '''
    
    def __init__(self, id, atoms = None):
        '''
        Parameters
        ----------
        id: str
            Metabolite ID.
        atoms: str or list of str or None
            Carbons in metabolite, e.g., 'abcd' for metabolite without equivalents, 
            and ['abcd', 'dcba'] for metabolite with equivalents.
        '''
        
        self.id = id
        self.atoms = atoms
        if self.atoms and isinstance(self.atoms, str):
            self.atoms = [self.atoms]
        self.host_reactions = None
        
    
    def __hash__(self):
        
        if self.atoms:
            return hash(self.id) + sum([hash(atoms) for atoms in self.atoms])
        
        else:
            return hash(self.id)
            
        
    def __eq__(self, other):
        '''
        Parameters
        ----------
        other: Metabolite
        '''
        
        if self.atoms:
            return self.id == other.id and set(self.atoms) == set(other.atoms)
        
        else:
            if other.atoms:
                return False
            else:
                return self.id == other.id
        
    
    @property
    def atoms_info(self):
    
        if self.atoms:
            return {atoms: 1/len(self.atoms) for atoms in self.atoms}
        
        else:
            return None
    
    
    @property
    def n_carbons(self):
        
        if self.atoms:
            return len(re.search(r'[a-z]+', self.atoms).group())
        
        else:
            return None
            
            
    def __repr__(self):
        
        atomStr = '(' + ','.join(self.atoms) + ')' if self.atoms else ''

        return f'{self.__class__.__name__} {self.id}{atomStr}'



#!/usr/bin/env python


from sys import version_info
from warnings import warn
from setuptools import setup


if version_info[:2] < (3, 7):
    warn('FreeFlux is not tested in early versions of Python.')

setup()



'''Define the EMU class.'''


__author__ = 'Chao Wu'


from functools import lru_cache
from collections.abc import Iterable
from .metabolite import Metabolite


class EMU():
    '''
    Define EMU (i.e., elementary metabolite unit) object and its operations.

    EMUs in the same metabolite and with the same atom NOs are considered as identical, 
    while metabolites which they derived from could be different.
    
    EMUs can be compared based self.metabolite_id and self.atom_nos.
    EMU and iterable object of EMUs can also be compared. In this case EMU will be put into
    the same iterable object with single item, and comparison between two iterables are performed.
    
    Currently only binary equivalents are considered.
    
    Parameters
    ----------
    id: str
        EMU ID
    metabolite: Metabolite or str
        Which metabolite the EMU comes from.
    atom_nos: list of int or str
        Atom NOs, sorted by number.

    Attributes
    ----------
    id: str
        EMU ID
    metabolite: Metabolite
        Which metabolite the EMU comes from.    
    metabolite_id: str
        Metabolite ID.
    atom_nos: list of int
        Atom NOs, sorted by number.
    size: int
        Size of EMU.
    equivalent_atom_nos: None or list of int
        Equivalent atom NOs, sorted by number.   
    equivalent: EMU
        Equivalent of EMU.
    '''
    
    def __init__(self, id, metabolite, atom_nos):
        '''
        Parameters
        ----------
        id: str
            EMU ID
        metabolite: Metabolite or str
            Which metabolite the EMU comes from.
        atom_nos: list of int or str
            Atom NOs, sorted by number.
        '''
        
        self.id = id
        if isinstance(metabolite, Metabolite):
            self.metabolite = metabolite
            self.metabolite_id = self.metabolite.id
        elif isinstance(metabolite, str):
            self.metabolite = Metabolite(metabolite)
            self.metabolite_id = metabolite
        if isinstance(atom_nos, list):
            self.atom_nos = sorted(atom_nos)
        elif isinstance(atom_nos, str):
            self.atom_nos = sorted(list(map(int, atom_nos)))
        self.size = len(self.atom_nos)
        
    
    def __hash__(self):
        
        return hash(self.metabolite_id) + hash(sum(self.atom_nos))
        
        
    def __eq__(self, other):
        '''
        Parameters
        ----------
        other: EMU or iterable
        '''
        
        if isinstance(other, Iterable):
            return type(other)([self]) == other
        else:
            return self.metabolite_id == other.metabolite_id and self.atom_nos == other.atom_nos
        
        
    def __lt__(self, other):
        '''
        Parameters
        ----------
        other: EMU or iterable
        '''
        
        if isinstance(other, Iterable):
            return type(other)([self]) < other
        else:
            if self.metabolite_id != other.metabolite_id:
                return self.metabolite_id < other.metabolite_id
            else:
                return self.atom_nos < other.atom_nos
                
    
    def __gt__(self, other):
        '''
        Parameters
        ----------
        other: EMU or iterable
        '''
        
        if isinstance(other, Iterable):
            return type(other)([self]) > other
        else:
            if self.metabolite_id != other.metabolite_id:
                return self.metabolite_id > other.metabolite_id
            else:
                return self.atom_nos > other.atom_nos
        
    
    @property
    @lru_cache()
    def equivalent_atom_nos(self):
        '''
        Returns
        -------
        equivAtomNOs: list of int or None
            Equivalent atom NOs, sorted by number.
        '''
        
        if len(self.metabolite.atoms_info) == 1:
            return None

        else:
            refAtoms, equivAtoms = self.metabolite.atoms_info   
            
            mapping = dict(zip(refAtoms, range(1, len(refAtoms)+1)))
            
            equivAtomNOs = sorted([mapping[equivAtoms[no-1]] for no in self.atom_nos])
            
            if equivAtomNOs == self.atom_nos:
                return None
            else:
                return equivAtomNOs
                
                
    @property
    @lru_cache()
    def equivalent(self):
        '''
        Returns
        -------
        EMU: EMU
            Equivalent of current EMU.
        '''
        
        equivAtomNOs = self.equivalent_atom_nos
        
        if equivAtomNOs:
            id = self.metabolite_id + '_' + ''.join(map(str, self.equivalent_atom_nos))
            metab = self.metabolite
            return EMU(id, metab, equivAtomNOs)
        else:
            return None
            
        
    def __repr__(self):
        
        return f'{self.__class__.__name__} {self.metabolite_id}_{"".join(map(str, self.atom_nos))}'
    
        


'''Define statistic functions.'''


__author__ = 'Chao Wu'


from os import makedirs
import numpy as np
import pandas as pd
from scipy.stats import t, chi2, probplot, zscore
from scipy.interpolate import interp1d
import matplotlib.pyplot as plt
import warnings
warnings.filterwarnings('ignore', category = RuntimeWarning)


def _chi2_test(obj_value, dof, confidence_level):
    '''
    Parameters
    ----------
    obj_value: float
        Objective value.
    dof: int
        Degree of freedom.
    confidence_level: float
        Confidence level, e.g., 0.95 as 95% confidence level.
    '''
    
    chi2Lb = chi2.ppf((1-confidence_level)/2, dof)
    chi2Ub = chi2.ppf((1+confidence_level)/2, dof)
    
    flag = '' if chi2Lb <= obj_value <= chi2Ub else 'not '
    
    print(
        f'SSR {obj_value:.2f} {flag}acceptable,'
        f' which is {flag}in [{round(chi2Lb, 2)}, {round(chi2Ub, 2)}]'
        f' of the {confidence_level*100}%% confidence level by chi2 test'
    )


def _normal_probability(resids, show_fig, output_dir):
    '''
    Parameters
    ----------
    resids: array
        Residuals.
    show_fig: bool
        Whether to show figure.
    output_dir: str
        Output directory.
    '''
    
    probplot(resids, plot = plt)
    ax = plt.gca()
    for item in [ax.xaxis.label, ax.yaxis.label] + ax.get_xticklabels() + ax.get_yticklabels():
        item.set_fontsize(20)
    ax.set_ylabel('Weighted residuals')
    ax.set_title('')
    
    if output_dir:
        makedirs(output_dir, exist_ok = True)
        plt.savefig(f'{output_dir}/normal_probability_plot.jpg', dpi = 300, bbox_inches = 'tight')
    
    if show_fig:
        plt.show()
    
    plt.close()


def _simulated_vs_measured_measurements(
        sim_meas, 
        exp_meas, 
        exp_meas_err, 
        xlabel, 
        ylabel, 
        xticklabels, 
        filename, 
        show_fig, 
        output_dir
):
    '''
    Parameters
    ----------
    sim_meas: array or list
        Simulated measurements.
    exp_meas: array or list
        Measured measurements.
    exp_meas_err: array or list
        Errors of measured measurements.
    xlabel: str
        Xlabel.
    ylabel: str
        Ylabel.
    xticklabels: array or list
        Xticklabels.
    filename: str
        File name.    
    show_fig: bool
        Whether to show figure.
    output_dir: str
        Output directory.
    '''
    
    sim_meas = np.array(sim_meas)
    exp_meas = np.array(exp_meas)
    exp_meas_err = np.array(exp_meas_err)
    
    barWidth = 0.4
    xPos = np.arange(1, sim_meas.size+1)
    
    plt.bar(
        x = xPos - 0.5*barWidth, 
        height = sim_meas, 
        width = barWidth, 
        label = 'simulated'
    )
    plt.bar(
        x = xPos + 0.5*barWidth, 
        height = exp_meas, 
        yerr = exp_meas_err, 
        width = barWidth, 
        label = 'measured', 
        capsize = 3
    )
    plt.xticks(ticks = xPos, labels = xticklabels, fontsize = 20)
    plt.tick_params(axis = 'y', labelsize = 20)
    plt.xlabel(xlabel, fontsize = 20)
    plt.ylabel(ylabel, fontsize = 20)
    
    plt.legend(
        loc = 'center', 
        bbox_to_anchor = (1.25, 0.5), 
        fontsize = 20, 
        frameon = False
    )
    
    if output_dir:
        makedirs(output_dir, exist_ok = True)
        #filename = xlabel if xlabel else ylabel
        plt.savefig(f'{output_dir}/{filename}.jpg', dpi = 300, bbox_inches = 'tight')
    
    if show_fig:
        plt.show()
    
    plt.close()
    
    
def _simulated_vs_measured_MDVs(simulated_MDVs, measured_MDVs, show_fig, output_dir):
    '''
    Parameters
    ----------
    simulated_MDVs: dict
        EMU ID => simulated MDV.
    measured_MDVs: dict
        EMU ID => [means, sds].
    show_fig: bool
        Whether to show figure.
    output_dir: str
        Output directory.    
    '''
    
    for emuid in measured_MDVs:
        simMDV = simulated_MDVs[emuid]
        expMDV, expMDVerr = measured_MDVs[emuid]
        xticklabels = [f'M{str(i)}' for i in range(simMDV.size)]
        _simulated_vs_measured_measurements(
            simMDV, 
            expMDV, 
            expMDVerr, 
            emuid, 
            'Isotopomer fraction', 
            xticklabels, 
            'sim_vs_exp_MDVs-'+emuid, 
            show_fig, 
            output_dir
        )
        
    
def _simulated_vs_measured_fluxes(simulated_fluxes, measured_fluxes, show_fig, output_dir):
    '''
    Parameters
    ----------
    simulated_fluxes: dict
        Flux ID => simulated flux.
    measured_fluxes: dict
        Flux ID => [mean, sd].
    show_fig: bool
        Whether to show figure.
    output_dir: str
        Output directory.
    '''
    
    simFluxes = []
    expFluxes = []
    expFluxesErr = []
    xticklabels = []
    for fluxid, simFlux in simulated_fluxes.items():
        expFlux, expFluxErr = measured_fluxes[fluxid]
        simFluxes.append(simFlux)
        expFluxes.append(expFlux)
        expFluxesErr.append(expFluxErr)
        xticklabels.append(fluxid)
    
    _simulated_vs_measured_measurements(
        simFluxes, 
        expFluxes, 
        expFluxesErr, 
        '', 
        'Flux', 
        xticklabels, 
        'sim_vs_exp_fluxes', 
        show_fig, 
        output_dir
    )


def _confidence_intervals_le(res, irr, cov, dof, confidence_level):
    '''
    Parameters
    ----------
    res: ser
        Optimal results, e.g., net fluxes, total fluxes or concentrations.
    irr: list
        Irreversible items. total fluxes and concentrations are all considered
        irreversible.
    cov: array
        Corvariance matrix of free fluxes obtained from hessian at convergence.
    dof: int
        Degree of freedom.
    confidence_level: float
        Confidence level, e.g., 0.95 as 95% confidence level.
    '''
    
    errors = t.ppf((1+confidence_level)/2, dof)/(dof+1)**0.5*np.diag(cov)**0.5
    lbs = res - errors
    ubs = res + errors
    
    ranges = {}
    for varid in res.index:
        if varid in irr:
            ranges[varid] = [max(0, lbs[varid]), ubs[varid]]
        else:
            ranges[varid] = [lbs[varid], ubs[varid]]
   
    return ranges
    

def _confidence_intervals_mc(res_set, irr, confidence_level):
    '''
    Parameters
    ----------
    res_set: list of ser
        Set of optimal net fluxes. Total fluxes or concentrations.
    irr: list
        Irreversible items. Total fluxes and concentrations are all considered
        irreversible.
    confidence_level: float
        Confidence level, e.g., 0.95 as 95% confidence level.
    '''
    
    resSet = pd.DataFrame(res_set)
    resSet = resSet[(np.abs(zscore(resSet)) < 3).all(axis = 1)]
    
    quantile = resSet.quantile([(1-confidence_level)/2, (1+confidence_level)/2])
    mask = resSet.apply(lambda col: col.between(*quantile[col.name]), axis = 0)
    selected = resSet[mask.all(axis = 1)]

    if selected.empty:
        raise ValueError('can not estimate CIs, need more Monte Carlo runs')
    else:
        ranges = {}
        for varid, col in selected.items():
            if varid in irr:
                ranges[varid] = [max(0, col.min()), col.max()]
            else:
                ranges[varid] = [col.min(), col.max()]
      
        return ranges


def _MDV_kinetics(emuid, simulated_inst_MDVs, show_fig, output_dir):
    '''
    Parameters
    ----------
    emuid: str
        EMU ID.
    simulated_inst_MDVs: dict
        Timepoint => MDV.
    show_fig: bool
        Whether to show figure.
    output_dir: str
        Output directory.
    '''
    
    tpoints = sorted(simulated_inst_MDVs)
    ts = np.linspace(tpoints[0], tpoints[-1], 100)
    mdvs = np.array(list(simulated_inst_MDVs.values()))
    interpMDVs = interp1d(tpoints, mdvs, axis = 0, kind = 1)(ts)
    
    for i in range(interpMDVs.shape[1]):
        plt.plot(ts, interpMDVs[:, i], label = 'M%s' % i, linewidth = 2)
        
    plt.xlabel('Time (s)', fontsize = 20)
    plt.ylabel('EMU %s' % emuid, fontsize = 20)
    plt.tick_params(labelsize = 20)
    plt.legend(
        loc = 'center', 
        bbox_to_anchor = (1.15, 0.5), 
        fontsize = 20, 
        frameon = False
    )
    
    if output_dir:
        makedirs(output_dir, exist_ok = True)
        plt.savefig(f'{output_dir}/{emuid}_kinetics.jpg', dpi = 300, bbox_inches = 'tight')
    
    if show_fig:
        plt.show()
    
    plt.close()
    
    
def _simulated_vs_measured_inst_MDVs(
        simulated_inst_MDVs, 
        measured_inst_MDVs, 
        show_fig, 
        output_dir
):
    '''
    Parameters
    ----------
    simulated_inst_MDVs: dict
        EMU ID => {t => simulated MDV}.
    measured_inst_MDVs: dict
        EMU ID => {t => [means, sds]}.
    show_fig: bool
        Whether to show figure.
    output_dir: str
        Output directory.
    '''
    
    for emuid in measured_inst_MDVs:
        simInstMDVs = simulated_inst_MDVs[emuid]
        expInstMDVs = measured_inst_MDVs[emuid]
    
        tpoints_sim = sorted(simInstMDVs)
        ts = np.linspace(tpoints_sim[0], tpoints_sim[-1], 100)
        mdvs = np.array(list(simInstMDVs.values()))
        interpMDVs = interp1d(tpoints_sim, mdvs, axis = 0, kind = 1)(ts)
        
        tpoint_exp = sorted(expInstMDVs)
        expMDVs = np.array([expInstMDVs[t][0] for t in tpoint_exp])
        expMDVerrs = np.array([expInstMDVs[t][1] for t in tpoint_exp])
        
        colors = []
        for i, simMDV in enumerate(interpMDVs.T):
            p = plt.plot(ts, simMDV, label = 'M%s' % i, linewidth = 2)
            colors.append(p[0].get_color())
        legend = plt.legend(
            loc = 'center', 
            bbox_to_anchor = (1.15, 0.5), 
            title = 'simulated', 
            title_fontsize = 20, 
            fontsize = 20, 
            frameon = False
        )
        plt.gca().add_artist(legend)    
        
        handles = []
        labels = []
        for i, (expMDV, expMDVerr) in enumerate(zip(expMDVs.T, expMDVerrs.T)):
            e = plt.errorbar(
                tpoint_exp, 
                expMDV, 
                expMDVerr, 
                color = colors[i], 
                linestyle = '',
                marker = '.', 
                markersize = 15, 
                elinewidth = 2, 
                capsize = 5
            )
            handles.append(e[0])
            labels.append('M%s' % i)
        plt.legend(
            handles = handles, 
            labels = labels, 
            loc = 'center', 
            bbox_to_anchor = (1.45, 0.5), 
            title = 'measured', 
            title_fontsize = 20, 
            fontsize = 20, 
            frameon = False
        )
            
        plt.xlabel('Time (s)', fontsize = 20)
        plt.ylabel(f'EMU {emuid}', fontsize = 20)
        plt.tick_params(labelsize = 20)
        
        if output_dir:
            makedirs(output_dir, exist_ok = True)
            plt.savefig(f'{output_dir}/{emuid}_inst.jpg', dpi = 300, bbox_inches = 'tight')
        
        if show_fig:
            plt.show()
        
        plt.close()
            

def _contribution_matrix(cov, trans_mat, simulated_der, measured_cov):
    '''
    Parameters
    ----------
    cov: array
        Corvariance matrix of free fluxes obtained from hessian at convergence.
    trans_mat: array
        Transformation matrix from free flux to total flux or net flux, i.e.,
        N to total flux, T@N to net flux.
    simulated_der: array
        Derivative of simulated measurements w.r.t. free fluxes.
    measured_cov: array
        Corvariance matrix of measurements.
    '''

    fluxesCov = trans_mat@cov@trans_mat.T
    fluxes_measCov = trans_mat@cov@simulated_der.T

    contribMat = np.empty_like(fluxes_measCov)
    for i in range(contribMat.shape[0]):
        for j in range(contribMat.shape[1]):
            contribMat[i,j] = fluxes_measCov[i,j]**2/(fluxesCov[i,i]*measured_cov[j,j])

    return contribMat


def _sensitivity(cov, trans_mat, simulated_der, measured_inv_cov):
    '''
    Parameters
    ----------
    cov: array
        Corvariance matrix of free fluxes obtained from hessian at convergence.
    trans_mat: array
        Transformation matrix from free flux to total flux or net flux, i.e.,
        N to total flux, T@N to net flux.
    simulated_der: array
        Derivative of simulated measurements w.r.t. free fluxes.
    measured_inv_cov: array
        Inversed corvariance matrix of measurements.
    '''

    return trans_mat@cov@simulated_der.T@measured_inv_cov



'''Define the Model class.'''


__author__ = 'Chao Wu'


import re
from functools import lru_cache, reduce
from collections import OrderedDict, deque
from collections.abc import Iterable
from itertools import chain
import numpy as np
import pandas as pd
from sympy import Integer
from multiprocessing import Pool
from .metabolite import Metabolite
from .reaction import Reaction
from .emu import EMU
from ..io.inputs import read_model_from_file
from ..analysis.simulate import Simulator
from ..analysis.inst_simulate import InstSimulator
from ..analysis.fit import Fitter
from ..analysis.inst_fit import InstFitter
from ..optim.optim import Optimizer


class Model():
    '''
    Model class is the fundamental class of FreeFlux hosting operations for 13C MFA.
    A model can be built by adding reactions one by one or reading set of reactions from 
    .tsv or .xlsx file.
    
    Parameters
    ----------
    name: str
        Model name.

    Attributes
    ----------
    name: str
        Model name.
    metabolites_info: dict
        Metabolite ID => list of Metabolites.
    reactions_info: OrderedDict
        Reaction ID => Reaction.
    metabolites: list
        Metabolite IDs, in alphabetical order.
    metabolites_with_atoms: list
        IDs of metabolite with atom assignment, in alphabetical order. 
    end_substrates: list
        Initial substrates of the model, in alphabetical order. 
    end_products: list
        Final products of the model, in alphabetical order. 
    reactions: list
        Reaction IDs, in order of addition.
    n_metabolites: int
        # of metabolites.
    n_reactions: int
        # of reactions.
    _full_net_stoichiometric_matrix: df
        Complete stoichiometric matrix for net reaction with all metabolites in rows, 
        net reactions in columns.
    _full_total_stoichiometric_matrix: df
        Complete stoichiometric matrix for total reaction with all metabolites in rows, 
        total reactions in columns.
    metabolite_adjacency_matrix: df
        Metabolite adjacency matrix (MAM). Metabolites with atoms are in index and 
        columns (no duplicates). List of Reactions are in cells if reactions exists 
        between (sub, pro), [] otherwise.    
    net_fluxes_bounds: dict
        Reaction ID => [lb, ub] by setting.   
    net_fluxes_range: dict
        Reaction ID => estimeted [lb, ub]. All required net fluxes are included.  
    netfluxids: list
        Net flux IDs, alias of reactions. 
    concentrations: ser
        Concentrations, metabolite ID => float.    
    concentrations_bounds: dict
        Metabolite ID => [lb, ub] by setting.    
    concentrations_range: dict
        Metabolite ID => [lb, ub]. All required concentrations are included.    
    concids: list
        Concentration IDs (used only in computation process).    
    total_fluxes: ser
        Total fluxes, fluxe ID (e.g., 'v1_f' or 'v2') => float.   
    totalfluxids: list
        Total flux IDs.    
    target_EMUs: list
        Target EMU IDs.   
    timepoints: list
        Sorted time points for MDV simulation.   
    substrate_MDVs: dict
        Substrate EMU => MDV.   
    substrate_MDVs_der_p: dict
        Substrate EMU => derivatives of substrate MDV w.r.t. variables 
        in shape of (len(MDV), # of vars).
        # of vars = # of free fluxes for steady state MFA;
        # of vars = # of free fluxes + # of concentrations for INST MFA.
    initial_matrix_Xs: dict
        Size => initial MDVs of EMU in matrix X, i.e., the natural MDVs.
        The initial MDV matrix has the same shape of matrix X.   
    initial_matrix_Ys: dict
        Size => initial MDVs of EMU in matrix Y, i.e., either natural MDVs or labeled MDVs.
        The initial MDV matrix has the same shape of matrix Y.        
    initial_matrix_Xs_der_p, initial_matrix_Ys_der_p: dict
        Size => 3-D array in shape of (# of vars), X(Y).shape[0], X(Y).shape[1])
        which is the initial MDV derivatives of EMUs in matrix X(Y) w.r.t. variables. 
        # of vars = # of free fluxes + # of concentrations for INST MFA.  
    initial_sim_MDVs: dict
        EMU ID => {t0 => MDV}. MDV of target EMUs at t0.
    EAMs: dict of df
        Size => EMU adjacency matrix (EAM). Cells are symbolic expression of flux.    
    matrix_As, matrix_Bs: dict
        Size => [lambdified matrix A(B), [flux IDs], [EMUs]].     
    matrix_As_der_p, matrix_Bs_der_p: dict
        Size => 3-D array in shape of (# of vars, A(B).shape[0], A(B).shape[1])
        which is the derivatives of matrix A(B) w.r.t. variables.
        # of vars = # of free fluxes for steady state MFA;
        # of vars = # of free fluxes + # of concentrations for INST MFA.    
    matrix_Ms: dict
        Size => [lambdified matrix M, [metabolite IDs]].    
    matrix_Ms_der_p: dict
        Size => 3-D array in shape of (# of vars, M.shape[0], M.shape[1]),
        which is the derivatives of matrix M w.r.t. variables.
        # of vars = # of free fluxes for steady state MFA;
        # of vars = # of free fluxes + # of concentrations for INST MFA.    
    label_atom: str
        Labeled atom, i.e., the base atom in MDV, currently only supprts "H", "C" and
        "N".
    labeling_strategy: dict
        Metabolite ID => [labeling_pattern(s), percentage(s), purity(s)].    
    measured_MDVs: dict
        EMU ID (metabolite ID + '_' + atom NOs) => [means of MDV, SDs of MDV].    
    measured_MDVs_inv_cov: array
        Inversed covariance matrix of measured MDVs with variances on the diagnol, 
        other elements are zero.    
    measured_fluxes: dict
        Irreversible reaction ID => [mean, SD].    
    measured_fluxes_inv_cov: array
        Inversed covariance matrix of measured fluxes with variances on the diagnol, 
        other elements are zero.    
    measured_fluxes_der_p: array
        Derivative of measured fluxes w.r.t. variables in shape of 
        (# of measured fluxes, # of vars),
        # of vars = # of free fluxes for steady state MFA;
        # of vars = # of free fluxes + # of concentrations for INST MFA.
    measured_inst_MDVs: dict
        EMU ID (metabolite ID + '_' + atom NOs) => {timepoint => [means of MDV, SDs of MDV]}.    
    measured_inst_MDVs_inv_cov: array
        Inversed covariance matrix of measured concatenated MDVs with variances on the diagnol, 
        other elements are zero. Timepoints are concatenated except t0.     
    null_space: 2-D array
        Null space of total stoichiometric matrix.    
    transform_matrix: 2-D array
        Transform matrix letting vnet = transform_matrix*v.
    '''
    
    def __init__(self, name = 'unnamed'):
        '''
        Parameters
        ----------
        name: str
            Model name.
        '''
        
        self.name = name
        self.reactions_info = OrderedDict()
        
        self.target_EMUs = []
        self.timepoints = []
        self.substrate_MDVs = {}
        self.substrate_MDVs_der_p = {}
        self.EAMs = {}
        self.matrix_As = {}
        self.matrix_Bs = {}
        self.matrix_Ms = {}
        self.matrix_As_der_p = {}
        self.matrix_Bs_der_p = {}
        self.matrix_Ms_der_p = {}
        
        self.null_space = None
        self.transform_matrix = None
        self.initial_matrix_Xs = {}
        self.initial_matrix_Ys = {}
        self.initial_matrix_Xs_der_p = {}
        self.initial_matrix_Ys_der_p = {}
        self.initial_sim_MDVs = {}
        
        self.label_atom = None
        self.labeling_strategy = {}
        self.measured_MDVs = {}
        self.measured_fluxes = {}
        self.measured_inst_MDVs = {}
        self.measured_MDVs_inv_cov = None
        self.measured_inst_MDVs_inv_cov = None
        self.measured_fluxes_inv_cov = None
        self.measured_fluxes_der_p = None
        self.unbalanced_metabolites = set()
        
        self.net_fluxes_bounds = {}
        self.net_fluxes_range = OrderedDict()
        self.concentrations_bounds = {}
        self.concentrations_range = OrderedDict()
        self.total_fluxes = pd.Series(dtype = float)
        self.concentrations = pd.Series(dtype = float)
            
        
    def add_reactions(self, reactions):
        '''
        Parameters
        ----------
        reactions: Reaction or list of Reaction
        '''
        
        if not isinstance(reactions, list):
            reactions = [reactions]
        
        new_reactions_info = OrderedDict([(rxn.id, rxn) for rxn in reactions])
        self.reactions_info.update(new_reactions_info)
        
        for rxn in reactions:
            if rxn.host_models is None:
                rxn.host_models = set([self])
            else:
                rxn.host_models.add(self)
    

    def remove_reactions(self, reactions):
        '''
        Parameters
        ----------
        reactions: Reaction or list of Reaction
        '''
        
        if not isinstance(reactions, list):
            reactions = [reactions]
            
        for rxn in reactions:
            self.reactions_info.pop(rxn.id)
            
        for rxn in reactions:
            rxn.host_models.remove(self)
            if not rxn.host_models:
                rxn.host_models = None
                break    
        
        
    def read_from_file(self, file):
        '''
        Parameters
        ----------
        file: file path
            tsv or excel file with reactions with fields "reaction_ID", "substrate_IDs(atom)", 
            "product_IDs(atom)" and "reversibility".

            Header line starts with "#", and will be skiped.
        '''
        
        dataRaw = read_model_from_file(file)
        
        data = pd.DataFrame()
        for col, ser in dataRaw.items():
            if ser.dtype == object:
                data[col] = ser.str.replace(r'\s+', '', regex = True)
            else:
                data[col] = ser
        
        pattern_re = re.compile(r'([0-9\.]+|)([.\w]+)\(?([a-z0-9\.,]+|)\)?')
        for rxn, (subsStr, prosStr, rev) in data.iterrows():
            
            v = Reaction(rxn, reversible = bool(int(rev)))
            
            for stoy, metab, atoms in pattern_re.findall(subsStr):
                stoy = float(stoy) if stoy else 1.0
                atoms = atoms.split(',') if atoms else None
                v.add_substrates(Metabolite(metab, atoms), stoy)
    
            for stoy, metab, atoms in pattern_re.findall(prosStr):
                stoy = float(stoy) if stoy else 1.0
                atoms = atoms.split(',') if atoms else None
                v.add_products(Metabolite(metab, atoms), stoy)
            
            self.add_reactions(v)
    
    
    @property
    def metabolites_info(self):
        '''
        Returns
        -------
        metabsInfo: dict
            metabolite ID => list of Metabolites
        '''
        
        metabsInfo = {}
        for _, rxn in self.reactions_info.items():
            for subid, sub in rxn.substrates_info['metab'].items():
                metabsInfo.setdefault(subid, []).append(sub)
            for proid, pro in rxn.products_info['metab'].items():
                metabsInfo.setdefault(proid, []).append(pro)
        
        metabsInfo = {metabid: list(set(metabs)) for metabid, metabs in metabsInfo.items()}
        
        return metabsInfo
        
    
    @property
    def metabolites(self):
        
        metabids = []
        for _, rxn in self.reactions_info.items():
            metabids = metabids + rxn.substrates + rxn.products
            
        return sorted(set(metabids))
    
    
    @property
    def metabolites_with_atoms(self):
        
        metabids = []
        for _, rxn in self.reactions_info.items():
            metabids = metabids + rxn.substrates_with_atoms + rxn.products_with_atoms
            
        return sorted(set(metabids))
    
    
    @property
    def reactions(self):
        
        return list(self.reactions_info.keys())
    
    
    @property
    def n_metabolites(self):
    
        return len(self.metabolites)
        
        
    @property
    def n_reactions(self):
    
        return len(self.reactions)
        
    
    @lru_cache()
    def _full_net_stoichiometric_matrix(self, metabolites, reactions):
        '''
        Re-estimate stoichiometric matrix if metabolites or reactions changes.

        Parameters
        ----------
        metabolites: tuple
            metabolite IDs
        reactions: tuple
            reaction IDs

        Returns
        -------
        netS: df
            Complete stoichiometric matrix for net reaction with all metabolites in rows and 
            net reactions in columns.
        '''
        
        netS = pd.DataFrame(0.0, index = metabolites, columns = reactions)
        for rxnid, rxn in self.reactions_info.items():
            
            for sub in rxn.substrates:
                value = rxn.substrates_info.loc[sub, 'stoy']
                if isinstance(value, pd.Series):
                    netS.loc[sub, rxnid] = -value.sum()   
                else:
                    netS.loc[sub, rxnid] = -value   
            
            for pro in rxn.products:
                value = rxn.products_info.loc[pro, 'stoy']
                if isinstance(value, pd.Series):
                    netS.loc[pro, rxnid] = value.sum()   
                else:
                    netS.loc[pro, rxnid] = value   
        
        return netS
    
    
    @lru_cache()
    def _full_total_stoichiometric_matrix(self, metabolites, reactions):
        '''
        Re-estimate stoichiometric matrix if metabolites or reactions changes.

        Parameters
        ----------
        metabolites: tuple
            metabolite IDs
        reactions: tuple
            reaction IDs

        Returns
        -------
        totalS: df
            Complete stoichiometric matrix for total reaction with all metabolites in rows and
            total reactions in columns.
        '''
        
        netS = self._full_net_stoichiometric_matrix(metabolites, reactions)
        
        totalS = []
        totalScols = []
        for rxn, col in netS.items():
            if self.reactions_info[rxn].reversible:
                totalS.append(col)
                totalS.append(-col+0.0)
                totalScols.append(rxn+'_f')
                totalScols.append(rxn+'_b')
            else:
                totalS.append(col)
                totalScols.append(rxn)

        totalS = pd.DataFrame(totalS, index = totalScols).T
        
        return totalS


    @property
    def end_substrates(self):
    
        totalS = self._full_total_stoichiometric_matrix(
            tuple(self.metabolites), 
            tuple(self.reactions)
        )
        
        nLessthan0 = totalS[totalS < 0].count(axis = 1)
        nGreaterthan0 = totalS[totalS > 0].count(axis = 1)
        
        return sorted(totalS.index[(nLessthan0 > 0) & (nGreaterthan0 == 0)].tolist())
        
        
    @property
    def end_products(self):
    
        totalS = self._full_total_stoichiometric_matrix(
            tuple(self.metabolites), 
            tuple(self.reactions)
        )
        
        nLessthan0 = totalS[totalS < 0].count(axis = 1)
        nGreaterthan0 = totalS[totalS > 0].count(axis = 1)
        
        return sorted(totalS.index[(nGreaterthan0 > 0) & (nLessthan0 == 0)].tolist())
        

    def get_net_stoichiometric_matrix(self, exclude_metabs = None, include_ends = False):
        '''
        Parameters
        ----------
        exclude_metabs: list or set
            Excluded metabolites.
        include_ends: bool
            Whether to include end metabolites (i.e., initial substrates and final products).
        
        Returns
        netS: df
            Net stoichiometric matrix with balanced metabolites in rows and net reactions in columns
        '''
        
        if exclude_metabs is None:
            exclude_metabs = []
        else:
            exclude_metabs = list(exclude_metabs)
            
        netS = self._full_net_stoichiometric_matrix(
            tuple(self.metabolites), 
            tuple(self.reactions)
        )
        
        if include_ends:
            selected_indices = netS.index.difference(exclude_metabs)
            return netS.loc[selected_indices,:]
        else:
            selected_indices = netS.index.difference(
                exclude_metabs+self.end_substrates+self.end_products
            )
            return netS.loc[selected_indices,:]
        
        
    def get_total_stoichiometric_matrix(self, exclude_metabs = None, include_ends = False):
        '''
        Parameters
        ----------
        exclude_metabs: list or set
            Excluded metabolites.
        include_ends: bool
            Whether to include end metabolites (initial substrates and final products).
        
        Returns
        totalS: df
            Total stoichiometric matrix with balanced metabolites in rows and total reactions 
            in columns
        '''
        
        if exclude_metabs is None:
            exclude_metabs = []
        else:
            exclude_metabs = list(exclude_metabs)    
        
        totalS = self._full_total_stoichiometric_matrix(
            tuple(self.metabolites), 
            tuple(self.reactions)
        )
        
        if include_ends:
            select_indices = totalS.index.difference(exclude_metabs)
            return totalS.loc[select_indices,:]
        else:
            select_indices = totalS.index.difference(
                exclude_metabs+self.end_substrates+self.end_products
            )
            return totalS.loc[select_indices,:]
    
    
    @property
    @lru_cache()
    def netfluxids(self):
        
        return self.reactions
        
        
    @property
    @lru_cache()
    def totalfluxids(self):
        
        totalfluxids = []
        for rxnid, rxn in self.reactions_info.items():
            if rxn.reversible:
                totalfluxids.extend([rxnid+'_f', rxnid+'_b'])
            else:
                totalfluxids.append(rxnid)
        
        return totalfluxids


    @property
    @lru_cache()
    def concids(self):
        
        if self.EAMs:
            metabids = []
            for _, EAM in self.EAMs.items():
                metabids += [emu.metabolite_id for emu in EAM.columns]
            return sorted(set(metabids))
        else:
            raise ValueError('network not decomposed')
        
    
    @property
    @lru_cache()
    def metabolite_adjacency_matrix(self):
        '''
        Returns
        -------
        MAM: df
            Metabolite adjacency matrix (MAM). Metabolites with atoms are in index and columns 
            (no duplicates). List of Reactions are in cells if reactions exists between sub (index) 
            and pro (columns), [] otherwise
        '''
        
        MAM = pd.DataFrame(
            index = self.metabolites_with_atoms, 
            columns = self.metabolites_with_atoms
        )
        MAM = MAM.applymap(lambda x: [])
        for _, rxn in self.reactions_info.items():
            for sub in rxn.substrates_with_atoms:
                for pro in rxn.products_with_atoms:
                    if rxn.reversible:
                        MAM.loc[sub, pro].append(rxn)
                        MAM.loc[pro, sub].append(rxn)
                    else:
                        MAM.loc[sub, pro].append(rxn)
        
        return MAM
    
    
    def __repr__(self):
    
        headStr = (
            f'{self.__class__.__name__} {self.name if self.name else "unknown"} '
            f'({self.n_metabolites} metabolites, {self.n_reactions} reactions)'
        )
        bodyStr = '\n'.join([str(rxn) for rxn in self.reactions_info.values()])

        return f'{headStr}\n{bodyStr}' 


    def _BFS(self, iniEMU):
        '''
        Parameters
        ----------
        iniEMU: EMU
            Starting EMU of the decomposition. 
            Metabolite of iniEMU can be any Metabolite instance with the same id.
        
        Returns
        -------
        EAMsInfo: dict of list
            Size => list of [EMU, [precursor EMUs], symbolic expression of flux].
        '''
        
        MAM = self.metabolite_adjacency_matrix
        
        EAMsInfo = {}
        
        searched = []
        toSearch = deque()
        toSearch.appendleft(iniEMU)
        while toSearch:
            
            currentEMU = toSearch.pop()
            searched.append(currentEMU)
            
            formingRxns = list(
                set(chain(*[cell for cell in MAM[currentEMU.metabolite_id] if cell]))
            )
            for formingRxn in formingRxns:
                
                if formingRxn.reversible:
                    if currentEMU.metabolite_id in formingRxn.products_with_atoms:
                        asProMetabs = formingRxn.products_info['metab'][currentEMU.metabolite_id]
                        direction = 'forward'
                        flux = formingRxn.fflux
                    else:
                        asProMetabs = formingRxn.substrates_info['metab'][currentEMU.metabolite_id]
                        direction = 'backward'
                        flux = formingRxn.bflux
                else:
                    asProMetabs = formingRxn.products_info['metab'][currentEMU.metabolite_id]
                    direction = 'forward'
                    flux = formingRxn.flux
                
                if isinstance(asProMetabs, pd.Series):
                    offset = 1/asProMetabs.size
                    asProMetabs = list(asProMetabs)
                else:
                    offset = 1.0
                    asProMetabs = [asProMetabs]
                
                
                for asProMetab in asProMetabs:    
                    currentEMU = EMU(currentEMU.id, asProMetab, currentEMU.atom_nos)   
                    preEMUsInfo = formingRxn._find_precursor_EMUs(currentEMU, direction = direction)
                    
                    for preEMUs, coe in preEMUsInfo:
                        for preEMU in preEMUs:
                            if preEMU not in searched and preEMU not in toSearch:
                                toSearch.appendleft(preEMU)
                            
                        EAMsInfo.setdefault(currentEMU.size, []).append(
                            [currentEMU, preEMUs, offset * coe * flux]
                        )
                            
        return EAMsInfo
    

    def _get_original_EAMs(self, iniEMU):
        '''
        Parameters
        ----------
        iniEMU: EMU
            Starting EMU of the decomposition.
            
        Returns
        -------
        EAMs: dict of df
            Size => original EMU adjacency matrix (EAM), cells are symbolic expression of flux。
        '''
        
        EAMsInfo = self._BFS(iniEMU)
        
        EAMs = {}
        for size, EMUsInfo in EAMsInfo.items():
            
            nonSourceEMUs = set([EMUInfo[0] for EMUInfo in EMUsInfo])
            SourceEMUs = sorted(
                set(
                    [tuple(EMUInfo[1]) if len(EMUInfo[1]) > 1 else EMUInfo[1][0] 
                     for EMUInfo in EMUsInfo]
                ) - nonSourceEMUs
            )
            
            EAM = pd.DataFrame(
                Integer(0), 
                index = sorted(nonSourceEMUs) + sorted(SourceEMUs), 
                columns = sorted(nonSourceEMUs)
            )
            for emu, preEMUs, flux in EMUsInfo:
                
                col = emu
                
                if len(preEMUs) == 1:
                    idx = preEMUs[0]
                else:
                    idx = tuple(preEMUs)
                
                EAM.loc[[idx], col] += flux
            
            EAMs[size] = EAM
        
        return EAMs
            
    
    def _replace_list_item(self, iterable, toReplace, value):
        '''
        Parameters
        ----------
        iterable: iterable
            Iterable can be nested.
        toReplace: scalar
            Value to be replaced.
        value: scalar
            Value that replaces toReplace.
            
        Returns
        -------
        newLst: tuple
            New tuple with toReplace recursively replaced.
        '''
        
        newLst = []
        for item in iterable:
            
            if not isinstance(item, Iterable):
                newItem = value if item == toReplace else item
            else:
                newItem = self._replace_list_item(item, toReplace, value)
            
            newLst.append(newItem)
        
        return tuple(newLst)
        
        
    def _uniquify_dataFrame_index(self, df):
        '''
        Parameters
        ----------
        df: df
            DataFrame to be uniquify.
            
        Returns
        -------
        uniqueDf: df
            DataFrame with duplicate rows combined (summated).
        '''
        
        sortedDf = df.sort_index()
    
        sortDfIndex, idx = np.unique(sortedDf.index, return_index = True)
    
        uniqueDf = pd.DataFrame(
            np.add.reduceat(sortedDf.values, idx), 
            index = sortDfIndex, 
            columns = sortedDf.columns
        )
        
        return uniqueDf
        
    
    def _lump_linear_EMUs(self, EAMs, iniEMU):
        '''
        Parameters
        ----------
        EAMs: dict of df
            Size => original EMU adjacency matrix (EAM), cells are symbolic expression of flux.
        iniEMU: EMU
            Starting EMU of the decomposition.
        
        Returns
        -------
        lumpedEAMs: dict of df
            Size => lumped EMU adjacency matrix (EAM), cells are symbolic expression of flux.
        '''
        
        lumpedEAMs = {}
        orderedSizes = sorted(EAMs.keys(), reverse = True)
        for i, size in enumerate(orderedSizes):
            
            lumpedEAM = EAMs[size].copy(deep = 'all')
            for emu in lumpedEAM.columns:
                
                preEMUs = lumpedEAM.index[lumpedEAM[emu] != 0]
                if preEMUs.size == 1:
                    preEMU = preEMUs[0]
                    
                    if (emu != iniEMU and 
                        (preEMU not in lumpedEAM.columns or lumpedEAM.loc[emu, preEMU] == 0)):
                        
                        lumpedEAM.drop(emu, axis = 1, inplace = True)
                        
                        lumpedEAM.index = self._replace_list_item(lumpedEAM.index, emu, preEMU)
                        
                        for j in range(i):
                            largerEAM = lumpedEAMs[orderedSizes[j]]
                            largerEAM.index = self._replace_list_item(largerEAM.index, emu, preEMU)
                        
                        lumpedEAM = self._uniquify_dataFrame_index(lumpedEAM)
                        upper = self._uniquify_dataFrame_index(
                            lumpedEAM.loc[lumpedEAM.columns, :]
                        )
                        lower = self._uniquify_dataFrame_index(
                            lumpedEAM.loc[lumpedEAM.index.difference(lumpedEAM.columns),:]
                        )
                        lumpedEAM = pd.concat((upper, lower))
            
            lumpedEAMs[size] = lumpedEAM
            
        return lumpedEAMs
        
        
    def _combine_equivalent_EMUs(self, EAMs):
        '''
        Parameters
        ----------
        EAMs: dict of df
            Size => original EMU adjacency matrix (EAM), cells are symbolic expression of flux.
            
        Returns
        -------
        combinedEAMs: dict df
            Size => EAM with equivalent EMUs combined, cells are symbolic expression of flux.
        '''
        
        combinedEAMs = {}
        for size, EAM in EAMs.items():
            
            combinedEAM = EAM.copy(deep = 'all')
            combined = []
            for emu in combinedEAM.columns:
                if emu not in combined:
                    
                    equivEMU = emu.equivalent
                    if equivEMU in combinedEAM.columns:
                        
                        combinedEAM.loc[:, emu] = combinedEAM.loc[:, [emu, equivEMU]].sum(axis = 1)/2
                        combinedEAM.drop(equivEMU, axis = 1, inplace = True)
                        
                        combinedEAM.loc[emu, :] = combinedEAM.loc[[emu, equivEMU], :].sum()
                        combinedEAM.drop(equivEMU, inplace = True)
                        
                        combined.append(equivEMU)
                    
            combinedEAMs[size] = combinedEAM

        return combinedEAMs
                    
        
    def get_emu_adjacency_matrices(self, iniEMU, lump = True):
        '''
        Parameters
        ----------
        iniEMU: EMU
            Starting EMU of the decomposition.
        lump: bool
            Whether to lump linear EMUs.
        
        Returns
        -------
        EAMs: dict of df
            Size => EMU adjacency matrix (EAM) after lumping of linear EMUs and combination
            of equivalent EMUs. Index and columns are EMUs, cells are symbolic expression of flux.

        Notes
        -----
        EMUs in sequential reactions can not be lumped in transient MFA.
        '''
        
        # set CPU affinity in Linux
        import platform
        if platform.system() == 'Linux':
            import os
            os.sched_setaffinity(os.getpid(), range(os.cpu_count()))
            
        oriEAMs = self._get_original_EAMs(iniEMU)
        
        if lump:
            lumpedEAMs = self._lump_linear_EMUs(oriEAMs, iniEMU)
            combinedEAMs = self._combine_equivalent_EMUs(lumpedEAMs)
        else:
            combinedEAMs = self._combine_equivalent_EMUs(oriEAMs)
        
        return combinedEAMs
        
    
    def _merge_EAMs(self, EAM1, EAM2):
        '''
        Parameters
        ----------
        EAM1: df
            EMU adjacency matrix (EAM) to merge
        EAM2: df
            EMU adjacency matrix (EAM) to merge
            
        Returns
        -------
        mergedEAM: df
            merged EAM
        '''
        
        nonSourceEMUunion = EAM2.columns.union(EAM1.columns)
        sourceEMUunion = EAM2.index.difference(EAM2.columns).union(
            EAM1.index.difference(EAM1.columns)
        )
        
        mergedEAM = pd.DataFrame(
            Integer(0), 
            index = nonSourceEMUunion.append(sourceEMUunion), 
            columns = nonSourceEMUunion
        )
        mergedEAM.loc[EAM1.index, EAM1.columns] = EAM1
        mergedEAM.loc[EAM2.index, EAM2.columns] = EAM2
        
        return mergedEAM
    
    
    def _merge_all_EAMs(self, *EAMsAll):
        '''
        Parameters
        ----------
        EAMsAll: tuple of EAMs
            EAMs is dict of DataFrame, i.e., Size => EMU adjacency matrix (EAM).

        Returns
        -------    
        mergedEAMs: dict of df
            Size => merged EAM     
        '''
        
        mergedEAMs = {}
        maxsize = max([max(EAMs) for EAMs in EAMsAll])
        for size in range(1, maxsize+1):
            EAMCurrentSize = list(
                filter(
                    lambda x: isinstance(x, pd.DataFrame), 
                    [EAMs.get(size, 0) for EAMs in EAMsAll]
                )
            )
            if EAMCurrentSize:
                mergedEAMs[size] = reduce(self._merge_EAMs, EAMCurrentSize)
        
        return mergedEAMs            
        
    
    def _decompose_network(self, metabolites, atom_nos, lump = True, n_jobs = 1):
        '''
        Parameters
        ----------
        metabolites: list of str
            List of metabolite IDs from which initial EMU will be generated to start the 
            decomposition.
        atom_nos: list of str
            Atom NOs of corresponding metabolites, len(atom_nos) should be equal to 
            len(metabolites).
        lump: bool
            Whether to lump linear EMUs.    
        n_jobs: int
            # of jobs to run in parallel.
        
        Returns
        -------
        mergedEAMs: dict of df
            Size => merged EMU adjacency matrix (EAM).

        Notes
        -----
        EMUs in sequential reactions can not be lumped in transient MFA.    
        '''
        
        emus = []
        for metabid, atomNOs in zip(metabolites, atom_nos):
            emus.append(EMU(metabid+'_'+atomNOs, Metabolite(metabid), atomNOs))
            
        EAMsAll = []
        if n_jobs == 1:
            for emu in emus:
                EAMs = self.get_emu_adjacency_matrices(emu, lump)
                EAMsAll.append(EAMs)
        else:
            pool = Pool(processes = n_jobs)
            
            for emu in emus:
                EAMs = pool.apply_async(
                    func = self.get_emu_adjacency_matrices, 
                    args = (emu, lump)
                )
                EAMsAll.append(EAMs)
            
            pool.close()    
            pool.join()
            
            EAMsAll = [EAMs.get() for EAMs in EAMsAll]
        
        mergedEAMs = self._merge_all_EAMs(*EAMsAll)
        
        return mergedEAMs    
        
        
    def decompose_network(self, ini_emus, lump = True, n_jobs = 1):
        '''
        Parameters
        ----------
        ini_emus: dict
            Metabolite ID => atom NOs or list of atom NOs. Atom NOs can be int list or str, 
            e.g., {'Ala': [[1,2,3], '23'], 'Ser': '123'}
        lump: bool
            Whether to lump linear EMUs.
        n_jobs: int
            # of jobs to run in parallel.
            
        Returns
        -------
        mergedEAMs: dict of df
            Size => merged EMU adjacency matrix (EAM)

        Notes
        -----
        EMUs in sequential reactions can not be lumped in transient MFA.      
        '''
        
        metabids = []
        atom_nos = []
        for metabid, atomNOs in ini_emus.items():
            if isinstance(atomNOs, list):
                if any(isinstance(item, Iterable) for item in atomNOs):
                    for atomnos in atomNOs:
                        if not isinstance(atomnos, str):
                            atomnos = ''.join(map(str, atomnos))
                        metabids.append(metabid)
                        atom_nos.append(atomnos)
                else:
                    atomNOs = ''.join(map(str, atomNOs))
                    metabids.append(metabid)
                    atom_nos.append(atomnos)
            else:
                metabids.append(metabid)
                atom_nos.append(atomNOs)            
        
        return self._decompose_network(metabids, atom_nos, lump = lump, n_jobs = n_jobs)        
    
    
    def optimizer(self):
    
        return Optimizer(self)
        
        
    def simulator(self, kind):
        '''
        Parameters
        ----------
        kind: {"ss", "inst"}
            * If "ss", simulation at isotopic steady state is performed.
            * If "inst", simulation at isotopically nonstationary state is performed.
        '''
        
        if kind.lower() == 'ss':
            return Simulator(self)
        elif kind.lower() == 'inst':
            return InstSimulator(self)
        else:
            raise ValueError('kind should be either "ss" or "inst"')
            
            
    def fitter(self, kind):
        '''
        Parameters
        ----------
        kind: {"ss", "inst"}
            * If "ss", fitting at isotopic steady state is performed.
            * If "inst", fitting at isotopically nonstationary state is performed.
        '''
        
        if kind.lower() == 'ss':
            return Fitter(self)
        elif kind.lower() == 'inst':
            return InstFitter(self)
        else:
            raise ValueError('kind should be either "ss" or "inst"')
